@@ -12,6 +12,7 @@ import (
 	"internal/buildcfg"
 	"internal/pkgbits"
 	"os"
+	"strings"
 
 	"cmd/compile/internal/base"
 	"cmd/compile/internal/ir"
@@ -62,9 +63,10 @@ import (
 type pkgWriter struct {
 	pkgbits.PkgEncoder
 
-	m      posMap
-	curpkg *types2.Package
-	info   *types2.Info
+	m                     posMap
+	curpkg                *types2.Package
+	info                  *types2.Info
+	rangeFuncBodyClosures map[*syntax.FuncLit]bool // non-public information, e.g., which functions are closures range function bodies?
 
 	// Indices for previously written syntax and types2 things.
 
@@ -89,13 +91,14 @@ type pkgWriter struct {
 
 // newPkgWriter returns an initialized pkgWriter for the specified
 // package.
-func newPkgWriter(m posMap, pkg *types2.Package, info *types2.Info) *pkgWriter {
+func newPkgWriter(m posMap, pkg *types2.Package, info *types2.Info, otherInfo map[*syntax.FuncLit]bool) *pkgWriter {
 	return &pkgWriter{
 		PkgEncoder: pkgbits.NewPkgEncoder(base.Debug.SyncFrames),
 
-		m:      m,
-		curpkg: pkg,
-		info:   info,
+		m:                     m,
+		curpkg:                pkg,
+		info:                  info,
+		rangeFuncBodyClosures: otherInfo,
 
 		pkgsIdx: make(map[*types2.Package]pkgbits.Index),
 		objsIdx: make(map[types2.Object]pkgbits.Index),
@@ -189,7 +192,9 @@ type writer struct {
 
 // A writerDict tracks types and objects that are used by a declaration.
 type writerDict struct {
-	implicits []*types2.TypeName
+	// implicits is a slice of type parameters from the enclosing
+	// declarations.
+	implicits []*types2.TypeParam
 
 	// derived is a slice of type indices for computing derived types
 	// (i.e., types that depend on the declaration's type parameters).
@@ -217,7 +222,7 @@ type itabInfo struct {
 // generic function or method.
 func (dict *writerDict) typeParamIndex(typ *types2.TypeParam) int {
 	for idx, implicit := range dict.implicits {
-		if types2.Unalias(implicit.Type()).(*types2.TypeParam) == typ {
+		if implicit == typ {
 			return idx
 		}
 	}
@@ -486,6 +491,18 @@ func (w *writer) typInfo(info typeInfo) {
 // typIdx also reports whether typ is a derived type; that is, whether
 // its identity depends on type parameters.
 func (pw *pkgWriter) typIdx(typ types2.Type, dict *writerDict) typeInfo {
+	// Strip non-global aliases, because they only appear in inline
+	// bodies anyway. Otherwise, they can cause types.Sym collisions
+	// (e.g., "main.C" for both of the local type aliases in
+	// test/fixedbugs/issue50190.go).
+	for {
+		if alias, ok := typ.(*types2.Alias); ok && !isGlobal(alias.Obj()) {
+			typ = alias.Rhs()
+		} else {
+			break
+		}
+	}
+
 	if idx, ok := pw.typsIdx[typ]; ok {
 		return typeInfo{idx: idx, derived: false}
 	}
@@ -498,7 +515,7 @@ func (pw *pkgWriter) typIdx(typ types2.Type, dict *writerDict) typeInfo {
 	w := pw.newWriter(pkgbits.RelocType, pkgbits.SyncTypeIdx)
 	w.dict = dict
 
-	switch typ := types2.Unalias(typ).(type) {
+	switch typ := typ.(type) {
 	default:
 		base.Fatalf("unexpected type: %v (%T)", typ, typ)
 
@@ -513,24 +530,20 @@ func (pw *pkgWriter) typIdx(typ types2.Type, dict *writerDict) typeInfo {
 
 		default:
 			// Handle "byte" and "rune" as references to their TypeNames.
-			obj := types2.Universe.Lookup(typ.Name())
+			obj := types2.Universe.Lookup(typ.Name()).(*types2.TypeName)
 			assert(obj.Type() == typ)
 
 			w.Code(pkgbits.TypeNamed)
-			w.obj(obj, nil)
+			w.namedType(obj, nil)
 		}
 
 	case *types2.Named:
-		obj, targs := splitNamed(typ)
-
-		// Defined types that are declared within a generic function (and
-		// thus have implicit type parameters) are always derived types.
-		if w.p.hasImplicitTypeParams(obj) {
-			w.derived = true
-		}
-
 		w.Code(pkgbits.TypeNamed)
-		w.obj(obj, targs)
+		w.namedType(splitNamed(typ))
+
+	case *types2.Alias:
+		w.Code(pkgbits.TypeNamed)
+		w.namedType(splitAlias(typ))
 
 	case *types2.TypeParam:
 		w.derived = true
@@ -571,7 +584,10 @@ func (pw *pkgWriter) typIdx(typ types2.Type, dict *writerDict) typeInfo {
 
 	case *types2.Interface:
 		// Handle "any" as reference to its TypeName.
-		if typ == anyTypeName.Type() {
+		// The underlying "any" interface is canonical, so this logic handles both
+		// GODEBUG=gotypesalias=1 (when any is represented as a types2.Alias), and
+		// gotypesalias=0.
+		if types2.Unalias(typ) == types2.Unalias(anyTypeName.Type()) {
 			w.Code(pkgbits.TypeNamed)
 			w.obj(anyTypeName, nil)
 			break
@@ -596,6 +612,17 @@ func (pw *pkgWriter) typIdx(typ types2.Type, dict *writerDict) typeInfo {
 	return typeInfo{idx: w.Flush(), derived: false}
 }
 
+// namedType writes a use of the given named type into the bitstream.
+func (w *writer) namedType(obj *types2.TypeName, targs *types2.TypeList) {
+	// Named types that are declared within a generic function (and
+	// thus have implicit type parameters) are always derived types.
+	if w.p.hasImplicitTypeParams(obj) {
+		w.derived = true
+	}
+
+	w.obj(obj, targs)
+}
+
 func (w *writer) structType(typ *types2.Struct) {
 	w.Len(typ.NumFields())
 	for i := 0; i < typ.NumFields(); i++ {
@@ -822,7 +849,11 @@ func (w *writer) doObj(wext *writer, obj types2.Object) pkgbits.CodeObj {
 	case *types2.TypeName:
 		if obj.IsAlias() {
 			w.pos(obj)
-			w.typ(obj.Type())
+			t := obj.Type()
+			if alias, ok := t.(*types2.Alias); ok { // materialized alias
+				t = alias.Rhs()
+			}
+			w.typ(t)
 			return pkgbits.ObjAlias
 		}
 
@@ -889,8 +920,7 @@ func (w *writer) objDict(obj types2.Object, dict *writerDict) {
 	// parameter is constrained to `int | uint` but then never used in
 	// arithmetic/conversions/etc, we could shape those together.
 	for _, implicit := range dict.implicits {
-		tparam := types2.Unalias(implicit.Type()).(*types2.TypeParam)
-		w.Bool(tparam.Underlying().(*types2.Interface).IsMethodSet())
+		w.Bool(implicit.Underlying().(*types2.Interface).IsMethodSet())
 	}
 	for i := 0; i < ntparams; i++ {
 		tparam := tparams.At(i)
@@ -1209,10 +1239,17 @@ func (w *writer) stmt(stmt syntax.Stmt) {
 func (w *writer) stmts(stmts []syntax.Stmt) {
 	dead := false
 	w.Sync(pkgbits.SyncStmts)
-	for _, stmt := range stmts {
-		if dead {
-			// Any statements after a terminating statement are safe to
-			// omit, at least until the next labeled statement.
+	var lastLabel = -1
+	for i, stmt := range stmts {
+		if _, ok := stmt.(*syntax.LabeledStmt); ok {
+			lastLabel = i
+		}
+	}
+	for i, stmt := range stmts {
+		if dead && i > lastLabel {
+			// Any statements after a terminating and last label statement are safe to omit.
+			// Otherwise, code after label statement may refer to dead stmts between terminating
+			// and label statement, see issue #65593.
 			if _, ok := stmt.(*syntax.LabeledStmt); !ok {
 				continue
 			}
@@ -1545,6 +1582,7 @@ func (w *writer) switchStmt(stmt *syntax.SwitchStmt) {
 	w.stmt(stmt.Init)
 
 	var iface, tagType types2.Type
+	var tagTypeIsChan bool
 	if guard, ok := stmt.Tag.(*syntax.TypeSwitchGuard); w.Bool(ok) {
 		iface = w.p.typeOf(guard.X)
 
@@ -1566,6 +1604,7 @@ func (w *writer) switchStmt(stmt *syntax.SwitchStmt) {
 			tv := w.p.typeAndValue(tag)
 			tagType = tv.Type
 			tagValue = tv.Value
+			_, tagTypeIsChan = tagType.Underlying().(*types2.Chan)
 		} else {
 			tagType = types2.Typ[types2.Bool]
 			tagValue = constant.MakeBool(true)
@@ -1618,12 +1657,18 @@ func (w *writer) switchStmt(stmt *syntax.SwitchStmt) {
 		// have the same type. If there are any case values that can't be
 		// converted to the tag value's type, then convert everything to
 		// `any` instead.
-	Outer:
-		for _, clause := range stmt.Body {
-			for _, cas := range syntax.UnpackListExpr(clause.Cases) {
-				if casType := w.p.typeOf(cas); !types2.AssignableTo(casType, tagType) {
-					tagType = types2.NewInterfaceType(nil, nil)
-					break Outer
+		//
+		// Except that we need to keep comparisons of channel values from
+		// being wrapped in any(). See issue #67190.
+
+		if !tagTypeIsChan {
+		Outer:
+			for _, clause := range stmt.Body {
+				for _, cas := range syntax.UnpackListExpr(clause.Cases) {
+					if casType := w.p.typeOf(cas); !types2.AssignableTo(casType, tagType) {
+						tagType = types2.NewInterfaceType(nil, nil)
+						break Outer
+					}
 				}
 			}
 		}
@@ -1659,7 +1704,11 @@ func (w *writer) switchStmt(stmt *syntax.SwitchStmt) {
 			w.Sync(pkgbits.SyncExprs)
 			w.Len(len(cases))
 			for _, cas := range cases {
-				w.implicitConvExpr(tagType, cas)
+				typ := tagType
+				if tagTypeIsChan {
+					typ = nil
+				}
+				w.implicitConvExpr(typ, cas)
 			}
 		}
 
@@ -2289,7 +2338,6 @@ func (w *writer) compLit(lit *syntax.CompositeLit) {
 				elem = kv.Value
 			}
 		}
-		w.pos(elem)
 		w.implicitConvExpr(elemType, elem)
 	}
 }
@@ -2302,6 +2350,7 @@ func (w *writer) funcLit(expr *syntax.FuncLit) {
 	w.Sync(pkgbits.SyncFuncLit)
 	w.pos(expr)
 	w.signature(sig)
+	w.Bool(w.p.rangeFuncBodyClosures[expr])
 
 	w.Len(len(closureVars))
 	for _, cv := range closureVars {
@@ -2362,12 +2411,16 @@ func (w *writer) varDictIndex(obj *types2.Var) {
 	}
 }
 
+// isUntyped reports whether typ is an untyped type.
 func isUntyped(typ types2.Type) bool {
-	basic, ok := types2.Unalias(typ).(*types2.Basic)
+	// Note: types2.Unalias is unnecessary here, since untyped types can't be aliased.
+	basic, ok := typ.(*types2.Basic)
 	return ok && basic.Info()&types2.IsUntyped != 0
 }
 
+// isTuple reports whether typ is a tuple type.
 func isTuple(typ types2.Type) bool {
+	// Note: types2.Unalias is unnecessary here, since tuple types can't be aliased.
 	_, ok := typ.(*types2.Tuple)
 	return ok
 }
@@ -2447,7 +2500,7 @@ type typeDeclGen struct {
 	gen int
 
 	// Implicit type parameters in scope at this type declaration.
-	implicits []*types2.TypeName
+	implicits []*types2.TypeParam
 }
 
 type fileImports struct {
@@ -2465,7 +2518,7 @@ type declCollector struct {
 	typegen    *int
 	file       *fileImports
 	withinFunc bool
-	implicits  []*types2.TypeName
+	implicits  []*types2.TypeParam
 }
 
 func (c *declCollector) withTParams(obj types2.Object) *declCollector {
@@ -2478,7 +2531,7 @@ func (c *declCollector) withTParams(obj types2.Object) *declCollector {
 	copy := *c
 	copy.implicits = copy.implicits[:len(copy.implicits):len(copy.implicits)]
 	for i := 0; i < n; i++ {
-		copy.implicits = append(copy.implicits, tparams.At(i).Obj())
+		copy.implicits = append(copy.implicits, tparams.At(i))
 	}
 	return &copy
 }
@@ -2573,6 +2626,10 @@ func (pw *pkgWriter) collectDecls(noders []*noder) {
 				pw.errorf(l.pos, "//go:linkname only allowed in Go files that import \"unsafe\"")
 				continue
 			}
+			if strings.Contains(l.remote, "[") && strings.Contains(l.remote, "]") {
+				pw.errorf(l.pos, "//go:linkname reference of an instantiation is not allowed")
+				continue
+			}
 
 			switch obj := pw.curpkg.Scope().Lookup(l.local).(type) {
 			case *types2.Func, *types2.Var:
@@ -2869,7 +2926,7 @@ func (pw *pkgWriter) isBuiltin(expr syntax.Expr, builtin string) bool {
 func recvBase(recv *types2.Var) *types2.Named {
 	typ := types2.Unalias(recv.Type())
 	if ptr, ok := typ.(*types2.Pointer); ok {
-		typ = ptr.Elem()
+		typ = types2.Unalias(ptr.Elem())
 	}
 	return typ.(*types2.Named)
 }
@@ -2913,6 +2970,9 @@ func objTypeParams(obj types2.Object) *types2.TypeParamList {
 		if !obj.IsAlias() {
 			return obj.Type().(*types2.Named).TypeParams()
 		}
+		if alias, ok := obj.Type().(*types2.Alias); ok {
+			return alias.TypeParams()
+		}
 	}
 	return nil
 }
@@ -2929,6 +2989,14 @@ func splitNamed(typ *types2.Named) (*types2.TypeName, *types2.TypeList) {
 	return typ.Obj(), typ.TypeArgs()
 }
 
+// splitAlias is like splitNamed, but for an alias type.
+func splitAlias(typ *types2.Alias) (*types2.TypeName, *types2.TypeList) {
+	orig := typ.Origin()
+	base.Assertf(typ.Obj() == orig.Obj(), "alias type %v has object %v, but %v has object %v", typ, typ.Obj(), orig, orig.Obj())
+
+	return typ.Obj(), typ.TypeArgs()
+}
+
 func asPragmaFlag(p syntax.Pragma) ir.PragmaFlag {
 	if p == nil {
 		return 0
