@@ -250,6 +250,7 @@ func writebarrier(f *Func) {
 		// to a new block.
 		var last *Value
 		var start, end int
+		var nonPtrStores int
 		values := b.Values
 	FindSeq:
 		for i := len(values) - 1; i >= 0; i-- {
@@ -261,8 +262,17 @@ func writebarrier(f *Func) {
 					last = w
 					end = i + 1
 				}
+				nonPtrStores = 0
 			case OpVarDef, OpVarLive:
 				continue
+			case OpStore:
+				if last == nil {
+					continue
+				}
+				nonPtrStores++
+				if nonPtrStores > 2 {
+					break FindSeq
+				}
 			default:
 				if last == nil {
 					continue
@@ -309,7 +319,7 @@ func writebarrier(f *Func) {
 						}
 
 						t := val.Type.Elem()
-						tmp := f.fe.Auto(w.Pos, t)
+						tmp := f.NewLocal(w.Pos, t)
 						mem = b.NewValue1A(w.Pos, OpVarDef, types.TypeMem, tmp, mem)
 						tmpaddr := b.NewValue2A(w.Pos, OpLocalAddr, t.PtrTo(), tmp, sp, mem)
 						siz := t.Size()
@@ -353,7 +363,7 @@ func writebarrier(f *Func) {
 		memThen := mem
 		var curCall *Value
 		var curPtr *Value
-		addEntry := func(v *Value) {
+		addEntry := func(pos src.XPos, v *Value) {
 			if curCall == nil || curCall.AuxInt == maxEntries {
 				t := types.NewTuple(types.Types[types.TUINTPTR].PtrTo(), types.TypeMem)
 				curCall = bThen.NewValue1(pos, OpWB, t, memThen)
@@ -394,7 +404,7 @@ func writebarrier(f *Func) {
 			val := w.Args[1]
 			if !srcs.contains(val.ID) && needWBsrc(val) {
 				srcs.add(val.ID)
-				addEntry(val)
+				addEntry(pos, val)
 			}
 			if !dsts.contains(ptr.ID) && needWBdst(ptr, w.Args[2], zeroes) {
 				dsts.add(ptr.ID)
@@ -407,7 +417,7 @@ func writebarrier(f *Func) {
 				// combine the read and the write.
 				oldVal := bThen.NewValue2(pos, OpLoad, types.Types[types.TUINTPTR], ptr, memThen)
 				// Save old value to write buffer.
-				addEntry(oldVal)
+				addEntry(pos, oldVal)
 			}
 			f.fe.Func().SetWBPos(pos)
 			nWBops--
@@ -449,6 +459,7 @@ func writebarrier(f *Func) {
 
 		// Do raw stores after merge point.
 		for _, w := range stores {
+			pos := w.Pos
 			switch w.Op {
 			case OpStoreWB:
 				ptr := w.Args[0]
@@ -483,6 +494,10 @@ func writebarrier(f *Func) {
 				mem.Aux = w.Aux
 			case OpVarDef, OpVarLive:
 				mem = bEnd.NewValue1A(pos, w.Op, types.TypeMem, w.Aux, mem)
+			case OpStore:
+				ptr := w.Args[0]
+				val := w.Args[1]
+				mem = bEnd.NewValue3A(pos, OpStore, types.TypeMem, w.Aux, ptr, val, mem)
 			}
 		}
 
@@ -656,7 +671,7 @@ func wbcall(pos src.XPos, b *Block, fn *obj.LSym, sp, mem *Value, args ...*Value
 	for i := 0; i < nargs; i++ {
 		argTypes[i] = typ
 	}
-	call := b.NewValue0A(pos, OpStaticCall, types.TypeResultMem, StaticAuxCall(fn, b.Func.ABIDefault.ABIAnalyzeTypes(nil, argTypes, nil)))
+	call := b.NewValue0A(pos, OpStaticCall, types.TypeResultMem, StaticAuxCall(fn, b.Func.ABIDefault.ABIAnalyzeTypes(argTypes, nil)))
 	call.AddArgs(args...)
 	call.AuxInt = int64(nargs) * typ.Size()
 	return b.NewValue1I(pos, OpSelectN, types.TypeMem, 0, call)
