@@ -29,10 +29,11 @@ import (
 	"cmd/go/internal/modload"
 	"cmd/go/internal/work"
 	"cmd/internal/quoted"
+	"cmd/internal/telemetry"
 )
 
 var CmdEnv = &base.Command{
-	UsageLine: "go env [-json] [-u] [-w] [var ...]",
+	UsageLine: "go env [-json] [-changed] [-u] [-w] [var ...]",
 	Short:     "print Go environment information",
 	Long: `
 Env prints Go environment information.
@@ -53,6 +54,10 @@ The -w flag requires one or more arguments of the
 form NAME=VALUE and changes the default settings
 of the named environment variables to the given values.
 
+The -changed flag prints only those settings whose effective
+value differs from the default value that would be obtained in
+an empty environment with no prior uses of the -w flag.
+
 For more about environment variables, see 'go help environment'.
 	`,
 }
@@ -64,19 +69,20 @@ func init() {
 }
 
 var (
-	envJson = CmdEnv.Flag.Bool("json", false, "")
-	envU    = CmdEnv.Flag.Bool("u", false, "")
-	envW    = CmdEnv.Flag.Bool("w", false, "")
+	envJson    = CmdEnv.Flag.Bool("json", false, "")
+	envU       = CmdEnv.Flag.Bool("u", false, "")
+	envW       = CmdEnv.Flag.Bool("w", false, "")
+	envChanged = CmdEnv.Flag.Bool("changed", false, "")
 )
 
 func MkEnv() []cfg.EnvVar {
-	envFile, _ := cfg.EnvFile()
+	envFile, envFileChanged, _ := cfg.EnvFile()
 	env := []cfg.EnvVar{
 		{Name: "GO111MODULE", Value: cfg.Getenv("GO111MODULE")},
-		{Name: "GOARCH", Value: cfg.Goarch},
+		{Name: "GOARCH", Value: cfg.Goarch, Changed: cfg.Goarch != runtime.GOARCH},
 		{Name: "GOBIN", Value: cfg.GOBIN},
-		{Name: "GOCACHE", Value: cache.DefaultDir()},
-		{Name: "GOENV", Value: envFile},
+		{Name: "GOCACHE"},
+		{Name: "GOENV", Value: envFile, Changed: envFileChanged},
 		{Name: "GOEXE", Value: cfg.ExeSuffix},
 
 		// List the raw value of GOEXPERIMENT, not the cleaned one.
@@ -90,62 +96,81 @@ func MkEnv() []cfg.EnvVar {
 		{Name: "GOHOSTARCH", Value: runtime.GOARCH},
 		{Name: "GOHOSTOS", Value: runtime.GOOS},
 		{Name: "GOINSECURE", Value: cfg.GOINSECURE},
-		{Name: "GOMODCACHE", Value: cfg.GOMODCACHE},
-		{Name: "GONOPROXY", Value: cfg.GONOPROXY},
-		{Name: "GONOSUMDB", Value: cfg.GONOSUMDB},
-		{Name: "GOOS", Value: cfg.Goos},
-		{Name: "GOPATH", Value: cfg.BuildContext.GOPATH},
+		{Name: "GOMODCACHE", Value: cfg.GOMODCACHE, Changed: cfg.GOMODCACHEChanged},
+		{Name: "GONOPROXY", Value: cfg.GONOPROXY, Changed: cfg.GONOPROXYChanged},
+		{Name: "GONOSUMDB", Value: cfg.GONOSUMDB, Changed: cfg.GONOSUMDBChanged},
+		{Name: "GOOS", Value: cfg.Goos, Changed: cfg.Goos != runtime.GOOS},
+		{Name: "GOPATH", Value: cfg.BuildContext.GOPATH, Changed: cfg.GOPATHChanged},
 		{Name: "GOPRIVATE", Value: cfg.GOPRIVATE},
-		{Name: "GOPROXY", Value: cfg.GOPROXY},
+		{Name: "GOPROXY", Value: cfg.GOPROXY, Changed: cfg.GOPROXYChanged},
 		{Name: "GOROOT", Value: cfg.GOROOT},
-		{Name: "GOSUMDB", Value: cfg.GOSUMDB},
+		{Name: "GOSUMDB", Value: cfg.GOSUMDB, Changed: cfg.GOSUMDBChanged},
 		{Name: "GOTMPDIR", Value: cfg.Getenv("GOTMPDIR")},
-		{Name: "GOTOOLCHAIN", Value: cfg.Getenv("GOTOOLCHAIN")},
+		{Name: "GOTOOLCHAIN"},
 		{Name: "GOTOOLDIR", Value: build.ToolDir},
 		{Name: "GOVCS", Value: cfg.GOVCS},
 		{Name: "GOVERSION", Value: runtime.Version()},
+		{Name: "GODEBUG", Value: os.Getenv("GODEBUG")},
+		{Name: "GOTELEMETRY", Value: telemetry.Mode()},
+		{Name: "GOTELEMETRYDIR", Value: telemetry.Dir()},
+	}
+
+	for i := range env {
+		switch env[i].Name {
+		case "GO111MODULE":
+			if env[i].Value != "on" && env[i].Value != "" {
+				env[i].Changed = true
+			}
+		case "GOBIN", "GOEXPERIMENT", "GOFLAGS", "GOINSECURE", "GOPRIVATE", "GOTMPDIR", "GOVCS":
+			if env[i].Value != "" {
+				env[i].Changed = true
+			}
+		case "GOCACHE":
+			env[i].Value, env[i].Changed = cache.DefaultDir()
+		case "GOTOOLCHAIN":
+			env[i].Value, env[i].Changed = cfg.EnvOrAndChanged("GOTOOLCHAIN", "")
+		case "GODEBUG":
+			env[i].Changed = env[i].Value != ""
+		}
 	}
 
 	if work.GccgoBin != "" {
-		env = append(env, cfg.EnvVar{Name: "GCCGO", Value: work.GccgoBin})
+		env = append(env, cfg.EnvVar{Name: "GCCGO", Value: work.GccgoBin, Changed: true})
 	} else {
 		env = append(env, cfg.EnvVar{Name: "GCCGO", Value: work.GccgoName})
 	}
 
-	key, val := cfg.GetArchEnv()
-	if key != "" {
-		env = append(env, cfg.EnvVar{Name: key, Value: val})
+	goarch, val, changed := cfg.GetArchEnv()
+	if goarch != "" {
+		env = append(env, cfg.EnvVar{Name: goarch, Value: val, Changed: changed})
 	}
 
 	cc := cfg.Getenv("CC")
+	ccChanged := true
 	if cc == "" {
+		ccChanged = false
 		cc = cfg.DefaultCC(cfg.Goos, cfg.Goarch)
 	}
 	cxx := cfg.Getenv("CXX")
+	cxxChanged := true
 	if cxx == "" {
+		cxxChanged = false
 		cxx = cfg.DefaultCXX(cfg.Goos, cfg.Goarch)
 	}
-	env = append(env, cfg.EnvVar{Name: "AR", Value: envOr("AR", "ar")})
-	env = append(env, cfg.EnvVar{Name: "CC", Value: cc})
-	env = append(env, cfg.EnvVar{Name: "CXX", Value: cxx})
+	ar, arChanged := cfg.EnvOrAndChanged("AR", "ar")
+	env = append(env, cfg.EnvVar{Name: "AR", Value: ar, Changed: arChanged})
+	env = append(env, cfg.EnvVar{Name: "CC", Value: cc, Changed: ccChanged})
+	env = append(env, cfg.EnvVar{Name: "CXX", Value: cxx, Changed: cxxChanged})
 
 	if cfg.BuildContext.CgoEnabled {
-		env = append(env, cfg.EnvVar{Name: "CGO_ENABLED", Value: "1"})
+		env = append(env, cfg.EnvVar{Name: "CGO_ENABLED", Value: "1", Changed: cfg.CGOChanged})
 	} else {
-		env = append(env, cfg.EnvVar{Name: "CGO_ENABLED", Value: "0"})
+		env = append(env, cfg.EnvVar{Name: "CGO_ENABLED", Value: "0", Changed: cfg.CGOChanged})
 	}
 
 	return env
 }
 
-func envOr(name, def string) string {
-	val := cfg.Getenv(name)
-	if val != "" {
-		return val
-	}
-	return def
-}
-
 func findEnv(env []cfg.EnvVar, name string) string {
 	for _, e := range env {
 		if e.Name == name {
@@ -205,7 +230,7 @@ func ExtraEnvVarsCostly() []cfg.EnvVar {
 		return q
 	}
 
-	return []cfg.EnvVar{
+	ret := []cfg.EnvVar{
 		// Note: Update the switch in runEnv below when adding to this list.
 		{Name: "CGO_CFLAGS", Value: join(cflags)},
 		{Name: "CGO_CPPFLAGS", Value: join(cppflags)},
@@ -215,6 +240,21 @@ func ExtraEnvVarsCostly() []cfg.EnvVar {
 		{Name: "PKG_CONFIG", Value: b.PkgconfigCmd()},
 		{Name: "GOGCCFLAGS", Value: join(cmd[3:])},
 	}
+
+	for i := range ret {
+		ev := &ret[i]
+		switch ev.Name {
+		case "GOGCCFLAGS": // GOGCCFLAGS cannot be modified
+		case "CGO_CPPFLAGS":
+			ev.Changed = ev.Value != ""
+		case "PKG_CONFIG":
+			ev.Changed = ev.Value != cfg.DefaultPkgConfig
+		case "CGO_CXXFLAGS", "CGO_CFLAGS", "CGO_FFLAGS", "GGO_LDFLAGS":
+			ev.Changed = ev.Value != work.DefaultCFlags
+		}
+	}
+
+	return ret
 }
 
 // argKey returns the KEY part of the arg KEY=VAL, or else arg itself.
@@ -296,27 +336,43 @@ func runEnv(ctx context.Context, cmd *base.Command, args []string) {
 	}
 
 	if len(args) > 0 {
-		if *envJson {
+		// Show only the named vars.
+		if !*envChanged {
+			if *envJson {
+				var es []cfg.EnvVar
+				for _, name := range args {
+					e := cfg.EnvVar{Name: name, Value: findEnv(env, name)}
+					es = append(es, e)
+				}
+				env = es
+			} else {
+				// Print just the values, without names.
+				for _, name := range args {
+					fmt.Printf("%s\n", findEnv(env, name))
+				}
+				return
+			}
+		} else {
+			// Show only the changed, named vars.
 			var es []cfg.EnvVar
 			for _, name := range args {
-				e := cfg.EnvVar{Name: name, Value: findEnv(env, name)}
-				es = append(es, e)
-			}
-			printEnvAsJSON(es)
-		} else {
-			for _, name := range args {
-				fmt.Printf("%s\n", findEnv(env, name))
+				for _, e := range env {
+					if e.Name == name {
+						es = append(es, e)
+						break
+					}
+				}
 			}
+			env = es
 		}
-		return
 	}
 
+	// print
 	if *envJson {
-		printEnvAsJSON(env)
-		return
+		printEnvAsJSON(env, *envChanged)
+	} else {
+		PrintEnv(os.Stdout, env, *envChanged)
 	}
-
-	PrintEnv(os.Stdout, env)
 }
 
 func runEnvW(args []string) {
@@ -422,12 +478,15 @@ func checkBuildConfig(add map[string]string, del map[string]bool) error {
 }
 
 // PrintEnv prints the environment variables to w.
-func PrintEnv(w io.Writer, env []cfg.EnvVar) {
+func PrintEnv(w io.Writer, env []cfg.EnvVar, onlyChanged bool) {
 	for _, e := range env {
 		if e.Name != "TERM" {
 			if runtime.GOOS != "plan9" && bytes.Contains([]byte(e.Value), []byte{0}) {
 				base.Fatalf("go: internal error: encountered null byte in environment variable %s on non-plan9 platform", e.Name)
 			}
+			if onlyChanged && !e.Changed {
+				continue
+			}
 			switch runtime.GOOS {
 			default:
 				fmt.Fprintf(w, "%s=%s\n", e.Name, shellQuote(e.Value))
@@ -502,12 +561,15 @@ func batchEscape(s string) string {
 	return b.String()
 }
 
-func printEnvAsJSON(env []cfg.EnvVar) {
+func printEnvAsJSON(env []cfg.EnvVar, onlyChanged bool) {
 	m := make(map[string]string)
 	for _, e := range env {
 		if e.Name == "TERM" {
 			continue
 		}
+		if onlyChanged && !e.Changed {
+			continue
+		}
 		m[e.Name] = e.Value
 	}
 	enc := json.NewEncoder(os.Stdout)
@@ -530,7 +592,7 @@ func checkEnvWrite(key, val string) error {
 	switch key {
 	case "GOEXE", "GOGCCFLAGS", "GOHOSTARCH", "GOHOSTOS", "GOMOD", "GOWORK", "GOTOOLDIR", "GOVERSION":
 		return fmt.Errorf("%s cannot be modified", key)
-	case "GOENV":
+	case "GOENV", "GODEBUG":
 		return fmt.Errorf("%s can only be set using the OS environment", key)
 	}
 
@@ -590,7 +652,7 @@ func checkEnvWrite(key, val string) error {
 }
 
 func readEnvFileLines(mustExist bool) []string {
-	file, err := cfg.EnvFile()
+	file, _, err := cfg.EnvFile()
 	if file == "" {
 		if mustExist {
 			base.Fatalf("go: cannot find go env config: %v", err)
@@ -654,7 +716,7 @@ func updateEnvFile(add map[string]string, del map[string]bool) {
 		}
 	}
 
-	file, err := cfg.EnvFile()
+	file, _, err := cfg.EnvFile()
 	if file == "" {
 		base.Fatalf("go: cannot find go env config: %v", err)
 	}
@@ -680,10 +742,7 @@ func lineToKey(line string) string {
 }
 
 // sortKeyValues sorts a sequence of lines by key.
-// It differs from sort.Strings in that keys which are GOx where x is an ASCII
-// character smaller than = sort after GO=.
-// (There are no such keys currently. It used to matter for GO386 which was
-// removed in Go 1.16.)
+// It differs from sort.Strings in that GO386= sorts after GO=.
 func sortKeyValues(lines []string) {
 	sort.Slice(lines, func(i, j int) bool {
 		return lineToKey(lines[i]) < lineToKey(lines[j])
