@@ -18,6 +18,8 @@ Flags:
 	-B note
 		Add an ELF_NT_GNU_BUILD_ID note when using ELF.
 		The value should start with 0x and be an even number of hex digits.
+		Alternatively, you can pass "gobuildid" in order to derive the
+		GNU build ID from the Go build ID.
 	-E entry
 		Set entry symbol name.
 	-H type
@@ -41,16 +43,22 @@ Flags:
 		or initialized to a constant string expression. -X will not work if the initializer makes
 		a function call or refers to other variables.
 		Note that before Go 1.5 this option took two separate arguments.
-	-a
-		Disassemble output.
 	-asan
 		Link with C/C++ address sanitizer support.
+	-aslr
+		Enable ASLR for buildmode=c-shared on windows (default true).
+	-bindnow
+		Mark a dynamically linked ELF object for immediate function binding (default false).
 	-buildid id
 		Record id as Go toolchain build id.
 	-buildmode mode
 		Set build mode (default exe).
 	-c
 		Dump call graphs.
+	-checklinkname=value
+		If value is 0, all go:linkname directives are permitted.
+		If value is 1 (the default), only a known set of widely-used
+		linknames are permitted.
 	-compressdwarf
 		Compress DWARF if possible (default true).
 	-cpuprofile file
@@ -62,8 +70,6 @@ Flags:
 		The dynamic header is on by default, even without any
 		references to dynamic libraries, because many common
 		system tools now assume the presence of the header.
-	-debugtramp int
-		Debug trampolines.
 	-dumpdep
 		Dump symbol dependency graph.
 	-extar ar
@@ -102,8 +108,6 @@ Flags:
 		Set runtime.MemProfileRate to rate.
 	-msan
 		Link with C/C++ memory sanitizer support.
-	-n
-		Dump symbol table.
 	-o file
 		Write output to file (default a.out, or a.out.exe on Windows).
 	-pluginpath path
@@ -114,13 +118,9 @@ Flags:
 		Link with race detection libraries.
 	-s
 		Omit the symbol table and debug information.
-	-shared
-		Generated shared object (implies -linkmode external; experimental).
 	-tmpdir dir
 		Write temporary files to dir.
 		Temporary files are only used in external linking mode.
-	-u
-		Reject unsafe packages.
 	-v
 		Print trace of linker operations.
 	-w
