@@ -19,3 +19,7 @@ func asmcgocall_no_g(fn, arg unsafe.Pointer)
 // respectively. Does not follow the Go ABI.
 func spillArgs()
 func unspillArgs()
+
+// getfp returns the frame pointer register of its caller or 0 if not implemented.
+// TODO: Make this a compiler intrinsic
+func getfp() uintptr { return 0 }
